"""
The order in which this tests are ran is important since resources are going to
be created by the tests in Apigee. A better way writing of the tests would be to
create a fixture that manages the creation and tear down of the relevant apps in
every tests but im a bit tired now...
"""
import pprint

import pytest

from pytest_nhsd_apim.apigee_apis import (
    AccessTokensAPI,
    ApigeeClient,
    ApigeeNonProdCredentials,
    ApiProductsAPI,
    AppKeysAPI,
    DebugSessionsAPI,
    DeveloperAppsAPI,
)


@pytest.fixture()
def client():
    config = ApigeeNonProdCredentials()
    return ApigeeClient(config=config)


class TestDeveloperAppsAPI:
    def test_create_app(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        body = {
            "apiProducts": ["identity-service-internal-dev"],
            "attributes": [
                {"name": "ADMIN_EMAIL", "value": "lucas.fantini@nhs.net"},
                {"name": "DisplayName", "value": "My App"},
                {"name": "Notes", "value": "Notes for developer app"},
                {"name": "MINT_BILLING_TYPE", "value": "POSTPAID"},
            ],
            "callbackUrl": "example.com",
            "name": "myapp",
            "scopes": [],
            "status": "approved",
        }
        pprint.pprint(
            developer_apps.create_app(email="lucas.fantini@nhs.net", body=body)
        )

    def test_post_app_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        body = {
            "apiProducts": ["identity-service-internal-dev"],
            "attributes": [
                {"name": "ADMIN_EMAIL", "value": "lucas.fantini@nhs.net"},
                {"name": "DisplayName", "value": "My App"},
                {"name": "Notes", "value": "Notes for developer app"},
                {"name": "MINT_BILLING_TYPE", "value": "POSTPAID"},
            ],
            "callbackUrl": "new-example.com",
            "name": "myapp",
            "scopes": [],
            "status": "approved",
        }
        pprint.pprint(
            developer_apps.post_app_by_name(
                email="lucas.fantini@nhs.net", body=body, app_name="myapp"
            )
        )

    def test_put_app_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        body = {
            "apiProducts": ["identity-service-internal-dev", "mock-jwks-internal-dev"],
            "attributes": [
                {"name": "ADMIN_EMAIL", "value": "lucas.fantini@nhs.net"},
                {"name": "DisplayName", "value": "My App"},
                {"name": "Notes", "value": "Notes for developer app"},
                {"name": "MINT_BILLING_TYPE", "value": "POSTPAID"},
            ],
            "callbackUrl": "new-example.com",
            "name": "myapp",
            "scopes": [],
            "status": "approved",
        }
        pprint.pprint(
            developer_apps.put_app_by_name(
                email="lucas.fantini@nhs.net", body=body, app_name="myapp"
            )
        )

    def test_list_apps(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(developer_apps.list_apps(email="lucas.fantini@nhs.net"))

    def test_get_app_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.get_app_by_name(
                email="lucas.fantini@nhs.net", app_name="myapp"
            )
        )

    def test_get_app_attributes(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.get_app_attributes(
                email="lucas.fantini@nhs.net", app_name="myapp"
            )
        )

    def test_post_app_attributes(self, client):
        body = {
            "attribute": [
                {"name": "ADMIN_EMAIL", "value": "admin@example.com"},
                {"name": "DisplayName", "value": "My App"},
            ]
        }
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.post_app_attributes(
                email="lucas.fantini@nhs.net", app_name="myapp", body=body
            )
        )

    def test_get_app_attributes_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.get_app_attribute_by_name(
                email="lucas.fantini@nhs.net",
                app_name="myapp",
                attribute_name="ADMIN_EMAIL",
            )
        )

    def test_post_app_attribute_by_name(self, client):
        body = {"name": "DisplayName", "value": "MyApp"}
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.post_app_attribute_by_name(
                email="lucas.fantini@nhs.net",
                app_name="myapp",
                attribute_name="DisplayName",
                body=body,
            )
        )

    def test_delete_app_attribute_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.delete_app_attribute_by_name(
                email="lucas.fantini@nhs.net",
                app_name="myapp",
                attribute_name="DisplayName",
            )
        )

    def test_delete_app_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.delete_app_by_name(
                email="lucas.fantini@nhs.net", app_name="myapp"
            )
        )


class TestApiProductsAPI:
    def test_get_apiproducts(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(api_products.get_products(expand=True))

    def test_post_apiproducts(self, client):
        api_products = ApiProductsAPI(client=client)
        body = {
            "apiResources": ["/"],
            "approvalType": "auto",
            "attributes": [{"name": "access", "value": "public"}],
            "description": "My API product",
            "displayName": "My API product",
            "environments": ["internal-dev"],
            "name": "myapiproduct",
            "proxies": ["identity-service-internal-dev"],
            "scopes": [],
        }
        pprint.pprint(api_products.post_products(body=body))

    def test_get_product_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(api_products.get_product_by_name(product_name="myapiproduct"))

    def test_put_product_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        body = {
            "apiResources": ["/"],
            "approvalType": "auto",
            "attributes": [{"name": "access", "value": "public"}],
            "description": "My API product",
            "displayName": "My API product",
            "environments": ["internal-dev"],
            "proxies": ["mock-jwks-internal-dev"],
            "scopes": [],
        }
        pprint.pprint(
            api_products.put_product_by_name(product_name="myapiproduct", body=body)
        )

    def test_get_product_attributes(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(api_products.get_product_attributes(product_name="myapiproduct"))

    def test_post_product_attributest(self, client):
        api_products = ApiProductsAPI(client=client)
        body = {"attribute": [{"name": "access", "value": "private"}]}
        pprint.pprint(
            api_products.post_product_attributest(
                product_name="myapiproduct", body=body
            )
        )

    def test_get_product_attribute_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(
            api_products.get_product_attribute_by_name(
                product_name="myapiproduct", attribute_name="access"
            )
        )

    def test_post_product_attribute_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        body = {"value": "new_value"}
        pprint.pprint(
            api_products.post_product_attribute_by_name(
                product_name="myapiproduct", attribute_name="access", body=body
            )
        )

    def test_delete_product_attribute_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(
            api_products.delete_product_attribute_by_name(
                product_name="myapiproduct", attribute_name="access"
            )
        )

    def test_delete_product_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(api_products.delete_product_by_name(product_name="myapiproduct"))


class TestDebugSessionAPI:
    def test_get_debugsession(self, client, _apigee_proxy):
        debugsession = DebugSessionsAPI(
            client=client,
            env_name=_apigee_proxy["environment"],
            api_name=_apigee_proxy["name"],
            revision_number=_apigee_proxy["revision"],
        )
        pprint.pprint(debugsession.get_debugsessions())

    def test_post_debugsession(self, client, _apigee_proxy):
        debugsession = DebugSessionsAPI(
            client=client,
            env_name=_apigee_proxy["environment"],
            api_name=_apigee_proxy["name"],
            revision_number=_apigee_proxy["revision"],
        )
        session = "my_session"
        pprint.pprint(debugsession.post_debugsession(session=session))

    def test_delete_debugsession_by_name(self, client, _apigee_proxy):
        debugsession = DebugSessionsAPI(
            client=client,
            env_name=_apigee_proxy["environment"],
            api_name=_apigee_proxy["name"],
            revision_number=_apigee_proxy["revision"],
        )

        pprint.pprint(
            debugsession.delete_debugsession_by_name(session_name="my_session")
        )

    def test_get_transaction_data(self, client, _apigee_proxy):
        debugsession = DebugSessionsAPI(
            client=client,
            env_name=_apigee_proxy["environment"],
            api_name=_apigee_proxy["name"],
            revision_number=_apigee_proxy["revision"],
        )
        debugsession.post_debugsession(session="my_session")
        pprint.pprint(debugsession.get_transaction_data(session_name="my_session"))
        debugsession.delete_debugsession_by_name(session_name="my_session")

    # get_transaction_data_by_id functionality needs a real test and is tested as part of test_trace in test_examples.py


class TestAccessTokenAPI:
    @pytest.mark.nhsd_apim_authorization({"access": "application", "level": "level3"})
    def test_get_token_details(self, client, _nhsd_apim_auth_token_data):
        # 1. Get an Apigee access_token
        token = _nhsd_apim_auth_token_data["access_token"]
        # 2. Get details
        token_api = AccessTokensAPI(client=client)
        pprint.pprint(token_api.get_token_details(access_token=token))

    @pytest.mark.nhsd_apim_authorization({"access": "application", "level": "level3"})
    def test_post_token_details(self, client, _nhsd_apim_auth_token_data):
        # 1. Get an Apigee access_token
        token = _nhsd_apim_auth_token_data["access_token"]
        # 2. Modify it
        token_api = AccessTokensAPI(client=client)
        body = {
            "scope": "my-malicious-scope",
            "attributes": [
                {"name": "my-malicious-name", "value": "my-malicious-value"}
            ],
        }  # scary...
        pprint.pprint(token_api.post_token_details(access_token=token, body=body))

    @pytest.mark.nhsd_apim_authorization({"access": "application", "level": "level3"})
    def test_delete_token(self, client, _nhsd_apim_auth_token_data):
        # 1. Get an Apigee access_token
        token = _nhsd_apim_auth_token_data["access_token"]
        # 2. Delete it
        token_api = AccessTokensAPI(client=client)
        pprint.pprint(token_api.delete_token(access_token=token))


class TestAppKeysAPI:
    def test_create_app(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        body = {
            "apiProducts": [],
            "attributes": [
                {"name": "ADMIN_EMAIL", "value": "lucas.fantini@nhs.net"},
                {"name": "DisplayName", "value": "My App"},
                {"name": "Notes", "value": "Notes for developer app"},
                {"name": "MINT_BILLING_TYPE", "value": "POSTPAID"},
            ],
            "callbackUrl": "example.com",
            "name": "myapp",
            "scopes": [],
            "status": "approved",
        }
        pprint.pprint(
            developer_apps.create_app(email="lucas.fantini@nhs.net", body=body)
        )

    def test_post_apiproducts(self, client):
        api_products = ApiProductsAPI(client=client)
        body = {
            "apiResources": ["/"],
            "approvalType": "auto",
            "attributes": [{"name": "access", "value": "public"}],
            "description": "My API product",
            "displayName": "My API product",
            "environments": ["internal-dev"],
            "name": "myapiproduct",
            "proxies": ["identity-service-internal-dev"],
            "scopes": ["scope1"],
        }
        pprint.pprint(api_products.post_products(body=body))

    def test_create_app_key(self, client):
        app_keys = AppKeysAPI(client=client)
        pprint.pprint(
            app_keys.create_app_key(
              "lucas.fantini@nhs.net",
              "myapp",
              {
                  "consumerKey": "RW3HpQ2oEGT8smAB",
                  "consumerSecret": "J0BUbo8Scy6M90qL"
              }
            )
        )

    def test_get_app_key(self, client):
        app_keys = AppKeysAPI(client=client)
        pprint.pprint(
            app_keys.get_app_key(
                "lucas.fantini@nhs.net",
                "myapp",
                "RW3HpQ2oEGT8smAB"
            )
        )

    def test_post_app_key(self, client):
        app_keys = AppKeysAPI(client=client)
        body = {
            "apiProducts": ["myapiproduct"]
        }
        pprint.pprint(
            app_keys.post_app_key(
                "lucas.fantini@nhs.net",
                "myapp",
                "RW3HpQ2oEGT8smAB",
                body
            )
        )

    def test_put_app_key(self, client):
        app_keys = AppKeysAPI(client=client)
        body = {
            "scopes": [
                "scope1"
            ]
        }
        pprint.pprint(
            app_keys.put_app_key(
                "lucas.fantini@nhs.net",
                "myapp",
                "RW3HpQ2oEGT8smAB",
                body
              )
        )

    def test_post_product_app_key_association(self, client):
        app_keys = AppKeysAPI(client=client)
        pprint.pprint(
            app_keys.post_product_app_key_association(
                "lucas.fantini@nhs.net",
                "myapp",
                "RW3HpQ2oEGT8smAB",
                "myapiproduct",
                action="revoke"
            )
        )
        pprint.pprint(
            app_keys.get_app_key(
                "lucas.fantini@nhs.net",
                "myapp",
                "RW3HpQ2oEGT8smAB"
            )
        )

    def test_post_app_key_params(self, client):
        app_keys = AppKeysAPI(client=client)
        pprint.pprint(
            app_keys.post_app_key(
                email="lucas.fantini@nhs.net",
                app_name="myapp",
                key="RW3HpQ2oEGT8smAB",
                body=None,
                action="revoke"
            )
        )
        pprint.pprint(
            app_keys.get_app_key(
                "lucas.fantini@nhs.net",
                "myapp",
                "RW3HpQ2oEGT8smAB"
            )
        )

    def test_delete_product_app_key_association(self, client):
        app_keys = AppKeysAPI(client=client)
        pprint.pprint(
            app_keys.delete_product_app_key_association(
                email="lucas.fantini@nhs.net",
                app_name="myapp",
                app_key="RW3HpQ2oEGT8smAB",
                apiproduct_name="myapiproduct"
            )
        )

    def test_delete_product_by_name(self, client):
        api_products = ApiProductsAPI(client=client)
        pprint.pprint(
            api_products.delete_product_by_name(
                product_name="myapiproduct"
            )
        )

    def test_delete_app_by_name(self, client):
        developer_apps = DeveloperAppsAPI(client=client)
        pprint.pprint(
            developer_apps.delete_app_by_name(
                email="lucas.fantini@nhs.net",
                app_name="myapp"
            )
        )


import pytest
import os


def test_examples(testdir):
    """Test the public fixtures."""

    testdir.copy_example("tests/test_examples.py")
    result = testdir.runpytest(
        "-vv",
        "-s",
    )

    # # fnmatch_lines does an assertion internally
    # result.stdout.fnmatch_lines([
    #     '*::test_sth PASSED*',
    # ])

    # make sure that we get a '0' exit code for the testsuite
    assert result.ret == 0


def test_internal_fixtures(testdir):
    """
    Test the private fixtures.
    """
    testdir.copy_example("tests/test_internals.py")
    result = testdir.runpytest(
        "-vv",
        "-s",
    )
    # make sure that we get a '0' exit code for the testsuite
    assert result.ret == 0



"""
This module defines all the pytest config.

- command line flags
- ini options
- custom markers

And a few fixtures to pull that config in.
"""
import os
from datetime import datetime

import pytest

from .log import log_method

_PYTEST_CONFIG = {
    "--api-name": {
        "help": "Name of API. Should match the meta/api/name field in your manifest file.",
        "default": "",  # Must be falsy but not None.
    },
    "--proxy-name": {
        "help": "Proxy under test, should exactly match the name on Apigee.",
        "default": "",  # Must be falsy but not None.
    },
    "--apigee-access-token": {"help": "Access token to log into apigee edge API, output of get_token"},
    "--apigee-organization": {
        "help": "nhsd-nonprod/nhsd-prod.",
        "default": "nhsd-nonprod",
    },
    "--apigee-developer": {
        "help": "Developer that will own the test app.",
        "default": "apm-testing-internal-dev@nhs.net",
    },
    "--apigee-app-id": {
        "help": "Apigee ID of application under test. Required for tests in production environments.",
        "default": "", # Must be falsy but not None.
    },
    "--status-endpoint-api-key": {
        "help": "Used to authenticate calls to proxy's _status endpoint. Required for tests in production environments.",
        "default": "",  # Must be falsy but not None.
    },
    "--jwt-public-key-id": {
        "help": "Key ID ('kid') to select particular key.",
        "default": "test-1",
    },
    # TODO: Add these if the user wishes...
    # "--jwt-public-key-url": {
    #     "help": "URL of JWT public key. Must be used with --jwt-private-key-file.",
    #     "default": None
    # },
    # "--jwt-private-key-file": {
    #     "help": "Path to private key of JWT. Must be used with --jwt-public-key-url.",
    #     "default": None,
    # }
}


def _flag_to_dest(flag):
    # e.g. --apigee-access-token is stored as APIGEE_ACCESS_TOKEN
    return flag[2:].replace("-", "_").upper()


def pytest_addoption(parser):
    """
    Hook for cli options.

    Pytest calls this at some point to get command line flags into the
    request.config object.

    I also want to be able to define sensitive config, e.g.
    apigee_access_token via environment variables.
    """

    group = parser.getgroup("nhsd-apim")
    for flag, attrs in _PYTEST_CONFIG.items():
        group.addoption(
            flag,
            action="store",
            dest=_flag_to_dest(flag),
            help=attrs.get("help"),
            default=attrs.get("default"),
        )


def pytest_configure(config):
    """
    Hook for defining markers.
    """
    config.addinivalue_line(
        "markers",
        f"nhsd_apim_authorization(*args, **kwargs): Marker to define the authorization journey you want to take to get your access token or apikey.",
    )


@pytest.fixture(scope="session")
@log_method
def nhsd_apim_config(request):
    """
    Use this fixture to access this pytest extension's config.
    It checks environment variables as well as the CLI.
    """

    def _get_config(flag):
        name = _flag_to_dest(flag)

        cmd_line_value = getattr(request.config.option, name)
        if cmd_line_value is not None:
            return cmd_line_value

        env_var_value = os.environ.get(name)
        if env_var_value is not None:
            return env_var_value

        ve_msg = f"Missing required config. You must pass cli option {flag} or environment variable {name}."
        raise ValueError(ve_msg)

    return {_flag_to_dest(flag): _get_config(flag) for flag in _PYTEST_CONFIG}


@pytest.fixture()
@log_method
def nhsd_apim_api_name(nhsd_apim_config, request):
    marker = request.node.get_closest_marker("nhsd_apim_api_name")
    if marker:
        return marker.args[0]
    if nhsd_apim_config["API_NAME"]:
        return nhsd_apim_config["API_NAME"]
    raise ValueError("API_NAME is not defined.")


@pytest.fixture()
@log_method
def nhsd_apim_proxy_name(nhsd_apim_config, request):
    marker = request.node.get_closest_marker("nhsd_apim_proxy_name")
    if marker:
        return marker.args[0]
    if nhsd_apim_config["PROXY_NAME"]:
        return nhsd_apim_config["PROXY_NAME"]
    raise ValueError("PROXY_NAME is not defined.")



"""
Making several API calls to get an access token per-test is very expensive.

This module is for caching access tokens.
"""
from typing import Optional, Dict, Any, Hashable, Tuple
from time import time
from functools import wraps

from .log import log, log_method


class _TokenCache:
    def __init__(self):
        self._cache = {}

    @log_method
    def insert(self, key: Hashable, token_data: Dict[str, Any]):
        if "issued_at" not in token_data:
            # only present on app-restricted tokens.  we can inject
            # this ourselves. Assume 5 seconds ago, probably was more
            # recently
            token_data["issued_at"] = time() - 5000
        self._cache[key] = token_data

    @log_method
    def get(self, key) -> Optional[Dict[str, Any]]:
        """
        Return a non-expired access token with the same client_id or None.
        """
        if key not in self._cache:
            return None

        old_token_data = self._cache[key]
        grace_period_seconds = 5
        now_ish = int(time()) + grace_period_seconds

        # issued_at is epoch_time in milliseconds
        # but expires_in is in seconds
        # => need factor of 1000 in this sum.
        expiry_time = int(old_token_data["issued_at"]) + 1000 * int(
            old_token_data["expires_in"]
        )
        if now_ish > expiry_time:
            self._cache.pop(key)
            return None
        return old_token_data


_CACHES = {}
_KWD_MARK = object()


def _recursive_make_hashable(d):
    if isinstance(d, dict):
        return tuple(
            sorted({k: _recursive_make_hashable(v) for k, v in d.items()}.items())
        )
    if isinstance(d, tuple):
        return tuple(_recursive_make_hashable(x) for x in d)

    return d


def cache_tokens(f):
    """
    Cache response tokens from an oauth server.

    This decorator checks a per-function _TokenCache for matching
    entries and returns the cached entries if they exist.

    The _TokenCache class understands that tokens have expiry times
    and automatically deletes them from the cache when appropriate.

    This means you can just call whatever get_access_token function
    you have decorated and not actually have to run the whole
    authentication journey.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        force_new_token = kwargs.pop("force_new_token", False)
        func_name = f.__name__
        if func_name not in _CACHES:
            _CACHES[func_name] = _TokenCache()
        cache = _CACHES[func_name]

        # We need to use the *args and **kwargs to construct a unique
        # dictionary key.  Dictionary keys must be hashable, which
        # means we can't have any dictionaries, but we can have
        # tuples. The tuple-of-just-a-blank-object, (_KWD_MARK,) is
        # how the python standard library marks the separation between
        # args and kwargs for its @functools.cache decorator. See
        # https://stackoverflow.com/a/10220908
        cache_key = (
            _recursive_make_hashable(args)
            + (_KWD_MARK,)
            + _recursive_make_hashable(kwargs)
        )
        token_data = cache.get(cache_key)
        if token_data and not force_new_token:
            log.debug(f"Cache hit for {func_name} with cache_key {cache_key}")
            return token_data
        log.debug(f"Cache miss for {func_name} with cache_key {cache_key}")
        token_data = f(*args, **kwargs)
        cache.insert(cache_key, token_data)
        return token_data

    return wrapper



"""
Example tests that show all the features of pytest_nhsd_apim.

These tests actually work!
"""
import json
import pytest
import requests

from uuid import uuid4

from pytest_nhsd_apim.identity_service import (
    ClientCredentialsConfig,
    ClientCredentialsAuthenticator,
    AuthorizationCodeConfig,
    AuthorizationCodeAuthenticator,
    KeycloakUserConfig,
    KeycloakUserAuthenticator,
    TokenExchangeConfig,
    TokenExchangeAuthenticator,
)


def test_ping_endpoint(nhsd_apim_proxy_url):
    """
    Send a request to an open access endpoint.
    """

    # The nhsd_apim_proxy_url will return the URL of the proxy under test.
    # The ping endpoint should have no authentication on it.

    resp = requests.get(nhsd_apim_proxy_url + "/_ping")
    assert resp.status_code == 200
    ping_data = json.loads(resp.text)
    assert "version" in ping_data


def test_status_endpoint(nhsd_apim_proxy_url, status_endpoint_auth_headers):
    """
    Send a request to the _status endpoint, protected by a platform-wide.
    """
    # The status_endpoint_auth_headers fixture returns a dict like
    # {"apikey": "thesecretvalue"} Use it to access your proxy's
    # _status endpoint.

    resp = requests.get(
        nhsd_apim_proxy_url + "/_status", headers=status_endpoint_auth_headers
    )
    status_json = resp.json()
    assert resp.status_code == 200
    assert status_json["status"] == "pass"


# Use the pytest.mark.nhsd_apim_authorization fixture automate the
# hassle of getting valid access tokens and/or API keys to access your
# API.
@pytest.mark.nhsd_apim_authorization({"access": "application", "level": "level0"})
def test_app_level0_access(nhsd_apim_proxy_url, nhsd_apim_auth_headers):
    """
    Test you have correctly configured an endpoint for application-level0 access.
    """
    print(nhsd_apim_auth_headers)
    resp = requests.get(nhsd_apim_proxy_url + "/test-auth/app/level0")
    assert resp.status_code == 401  # unauthorized

    resp = requests.get(
        nhsd_apim_proxy_url + "/test-auth/app/level0", headers=nhsd_apim_auth_headers
    )
    assert resp.status_code == 200  # authorized


# You can provide arguments as a dict (as above) or keyword-args.
# With these arguments, you get a public/private key pair, the public
# key is "hosted" on our mock-jwks proxy, and the
# nhsd_apim_auth_headers fixture does the signed JWT flow for you.
@pytest.mark.nhsd_apim_authorization(access="application", level="level3")
def test_app_level3_access(nhsd_apim_proxy_url, nhsd_apim_auth_headers):
    """
    Test you have correctly configured an endpoint for application-level3 access.
    """
    resp = requests.get(nhsd_apim_proxy_url + "/test-auth/app/level3")
    assert resp.status_code == 401  # unauthorized

    resp = requests.get(
        nhsd_apim_proxy_url + "/test-auth/app/level3", headers=nhsd_apim_auth_headers
    )
    assert resp.status_code == 200  # authorized


# All the flavours of access tokens you can request via this pytest
# extension are cached internally. This makes running your tests as
# fast as possible. In this example you are actually using the same
# access token as the one you obtained in the previous test!
@pytest.mark.parametrize(("count"), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
@pytest.mark.nhsd_apim_authorization({"access": "application", "level": "level3"})
def test_app_level3_access_repeatedly(count, nhsd_apim_auth_headers):
    def th(i):
        if i == 1:
            return "st"
        if i == 2:
            return "nd"
        if i == 3:
            return "rd"
        return "th"

    print(
        f"This is the {count}{th(count)} test using the same credentials - {nhsd_apim_auth_headers}"
    )


# If for any reason you want to override the caching
# the force_new_token flag can be added
@pytest.mark.parametrize(("count"), [1, 2, 3])
@pytest.mark.nhsd_apim_authorization(
    {"access": "application", "level": "level3", "force_new_token": True}
)
def test_app_level3_with_force_new_token(count, nhsd_apim_auth_headers):
    def th(i):
        if i == 1:
            return "st"
        if i == 2:
            return "nd"
        if i == 3:
            return "rd"
        return "th"

    print(
        f"This is the {count}{th(count)} test using different credentials - {nhsd_apim_auth_headers}"
    )


# You can include marks in a pytest parametrization to reduce
# boiler plate. This simple example matches authorization headers to
# paths and status codes. Once again, access tokens are cached for as
# long as they are valid so there are no unnecessary calls to the
# oauth server.
@pytest.mark.parametrize(
    ("expected_status_code", "test_path"),
    [
        pytest.param(
            200,
            "/test-auth/nhs-login/P0",
            marks=pytest.mark.nhsd_apim_authorization(
                access="patient",
                level="P0",
                login_form={"username": "9912003073"},
            ),
        ),
        pytest.param(
            403,
            "/test-auth/nhs-login/P5",
            marks=pytest.mark.nhsd_apim_authorization(
                access="patient",
                level="P0",
                login_form={"username": "9912003073"},
            ),
        ),
        pytest.param(
            403,
            "/test-auth/nhs-login/P9",
            marks=pytest.mark.nhsd_apim_authorization(
                access="patient",
                level="P0",
                login_form={"username": "9912003073"},
            ),
        ),
        pytest.param(
            200,
            "/test-auth/nhs-login/P5",
            marks=pytest.mark.nhsd_apim_authorization(
                access="patient",
                level="P5",
                login_form={"username": "9912003072"},
            ),
        ),
        pytest.param(
            403,
            "/test-auth/nhs-login/P9",
            marks=pytest.mark.nhsd_apim_authorization(
                access="patient",
                level="P5",
                login_form={"username": "9912003072"},
            ),
        ),
        pytest.param(
            200,
            "/test-auth/nhs-login/P9",
            marks=pytest.mark.nhsd_apim_authorization(
                access="patient",
                level="P9",
                login_form={"username": "9912003071"},
            ),
        ),
    ],
)
def test_patient_access_level_with_parametrization(
    nhsd_apim_proxy_url, expected_status_code, test_path, nhsd_apim_auth_headers
):
    """
    This parametrized test allows us to quickly check that each path
    gives us an expected status code depending on the details of how
    we authenticated to get our access token.
    """
    resp = requests.get(nhsd_apim_proxy_url + test_path, headers=nhsd_apim_auth_headers)
    assert resp.status_code == expected_status_code


# We are on our second generation of mock identity provider for
# healthcare_worker access (CIS2). This allows you to log-in using a
# username.
MOCK_CIS2_USERNAMES = {
    "aal1": ["656005750110"],
    "aal2": ["656005750109", "656005750111", "656005750112"],
    "aal3": ["656005750104", "656005750105", "656005750106"],
}


# Create a list of pytest.param for each combination of username and level for combined auth
combined_auth_params = [
    pytest.param(
        username, level,
        marks=pytest.mark.nhsd_apim_authorization(
            access="healthcare_worker",
            level=level,
            login_form={"username": username},
        ),
    )
    for level, usernames in MOCK_CIS2_USERNAMES.items()
    for username in usernames
]


@pytest.mark.parametrize("username, level", combined_auth_params)
def test_healthcare_worker_user_restricted_combined_auth(
    nhsd_apim_proxy_url, nhsd_apim_auth_headers, username, level
):
    url = f"{nhsd_apim_proxy_url}/test-auth/nhs-cis2/{level}"
    resp0 = requests.get(url)
    assert resp0.status_code == 401
    resp1 = requests.get(url, headers=nhsd_apim_auth_headers)
    assert resp1.status_code == 200


# Second generation auth also allows us to simulate separate authentication and
# authorization, also called "token exchange". In this flow, we authenticate
# directly with our Mock CIS2 instance, get an ID token, and exchange it via a
# call to our oauth server for an NHSD APIM access token. From out here, it
# doesn't look too different. "combined" authentication is the default for this
# library. To use separate authentication instead, add authentication="separate"
# to the nhsd_apim_authorization mark.

# Create a combined list for separate authentication testing
separate_auth_params = [
    pytest.param(
        username, level,
        marks=pytest.mark.nhsd_apim_authorization(
            access="healthcare_worker",
            level=level,
            login_form={"username": username},
            authentication="separate",
        ),
    )
    for level, usernames in MOCK_CIS2_USERNAMES.items()
    for username in usernames
]


@pytest.mark.parametrize("username, level", separate_auth_params)
def test_healthcare_work_user_restricted_separate_auth(
    nhsd_apim_proxy_url, nhsd_apim_auth_headers, username, level
):
    aal_url = f"{nhsd_apim_proxy_url}/test-auth/nhs-cis2/{level}"
    resp0 = requests.get(aal_url)
    assert resp0.status_code == 401
    resp1 = requests.get(aal_url, headers=nhsd_apim_auth_headers)
    assert resp1.status_code == 200


# we can also authenticate directly with our Mock NHS-Login instance, get an ID
# token, and exchange it via a call to our oauth server for an NHSD APIM access
# token. To use separate authentication instead, add authentication="separate"
# to the nhsd_apim_authorization mark.
@pytest.mark.nhsd_apim_authorization(
    {
        "access": "patient",
        "level": "P9",
        "login_form": {"username": "9912003071"},
        "authentication": "separate",
    }
)
def test_patient_user_restricted_separate_auth(
    nhsd_apim_proxy_url, nhsd_apim_auth_headers
):
    P9_url = f"{nhsd_apim_proxy_url}/test-auth/nhs-login/P9"
    resp0 = requests.get(P9_url)
    assert resp0.status_code == 401
    resp1 = requests.get(P9_url, headers=nhsd_apim_auth_headers)
    assert resp1.status_code == 200


# You can test unauthenticated access with an empty
# nhsd_apim_authorization marker.  This allows you to parameterize
# unauthenicated access tests in the same way as authenticated API calls.
@pytest.mark.nhsd_apim_authorization()
def test_no_authorization_explicitly_marked(
    nhsd_apim_proxy_url, nhsd_apim_auth_headers
):
    assert nhsd_apim_auth_headers == {}


# You can also do it without marking, though this raises a warning.
def test_no_authorization_with_not_explicitly_marked(
    nhsd_apim_proxy_url, nhsd_apim_auth_headers
):
    assert nhsd_apim_auth_headers == {}


# You can also use the authenticators directly in case you want to run the tests
# using a specific app already available in Apigee, leaving all the marker magic
# behind this is how you can implement the diferent authenticators
def test_client_credentials_authenticator(
    _test_app_credentials, _jwt_keys, apigee_environment
):
    # 1. Set your app config
    config = ClientCredentialsConfig(
        environment=apigee_environment,
        identity_service_base_url=f"https://{apigee_environment}.api.service.nhs.uk/oauth2-mock",
        client_id=_test_app_credentials["consumerKey"],
        jwt_private_key=_jwt_keys["private_key_pem"],
        jwt_kid="test-1",
    )

    # 2. Pass the config to the Authenticator
    authenticator = ClientCredentialsAuthenticator(config=config)

    # 3. Get your token
    token_response = authenticator.get_token()
    assert "access_token" in token_response
    token = token_response["access_token"]

    # 4. Use the token and confirm is valid
    headers = {"Authorization": f"Bearer {token}"}
    resp = requests.get(
        f"https://{apigee_environment}.api.service.nhs.uk/mock-jwks/test-auth/app/level3",
        headers=headers,
    )
    assert resp.status_code == 200


def test_authorization_code_authenticator(_test_app_credentials, apigee_environment):
    # 1. Set your app config
    config = AuthorizationCodeConfig(
        environment=apigee_environment,
        identity_service_base_url=f"https://{apigee_environment}.api.service.nhs.uk/oauth2-mock",
        callback_url="https://example.org/callback",
        client_id=_test_app_credentials["consumerKey"],
        client_secret=_test_app_credentials["consumerSecret"],
        scope="nhs-cis2",
        login_form={"username": "656005750104"},
    )

    # 2. Pass the config to the Authenticator
    authenticator = AuthorizationCodeAuthenticator(config=config)

    # 3. Get your token
    token_response = authenticator.get_token()
    assert "access_token" in token_response
    token = token_response["access_token"]

    # 4. Use the token and confirm is valid
    headers = {"Authorization": f"Bearer {token}"}
    resp = requests.get(
        f"https://{apigee_environment}.api.service.nhs.uk/mock-jwks/test-auth/nhs-cis2/aal3",
        headers=headers,
    )
    assert resp.status_code == 200


def test_token_exchange_authenticator(
    _test_app_credentials, apigee_environment, _jwt_keys, _keycloak_client_credentials
):
    # This one is a bit more interesting since we first need to get a token from
    # keycloak (our mock version of the NHS-Login provider) grab the id_token
    # from that token and then make a call to the identity service to "exchange"
    # the id_token for an Apigee access token.

    # 1. Set yout keycloack config
    config1 = KeycloakUserConfig(
        realm=f"NHS-Login-mock-{apigee_environment}",
        client_id=_keycloak_client_credentials["nhs-login"]["client_id"],
        client_secret=_keycloak_client_credentials["nhs-login"]["client_secret"],
        login_form={"username": "9912003071"},
    )

    # 2. Pass the config to the Authenticator
    authenticator = KeycloakUserAuthenticator(config=config1)

    # 3. Get your id_token
    id_token = authenticator.get_token()["id_token"]

    # 4. Set yout token exchange config
    config = TokenExchangeConfig(
        environment=apigee_environment,
        identity_service_base_url=f"https://{apigee_environment}.api.service.nhs.uk/oauth2-mock",
        client_id=_test_app_credentials["consumerKey"],
        jwt_private_key=_jwt_keys["private_key_pem"],
        jwt_kid="test-1",
        id_token=id_token,
    )

    # 5. Pass the config to the Authenticator
    authenticator = TokenExchangeAuthenticator(config=config)

    # 6. Get your token
    token_response = authenticator.get_token()
    assert "access_token" in token_response
    token = token_response["access_token"]

    # 7. Use the token and confirm is valid
    headers = {"Authorization": f"Bearer {token}"}
    resp = requests.get(
        f"https://{apigee_environment}.api.service.nhs.uk/mock-jwks/test-auth/nhs-login/P9",
        headers=headers,
    )
    assert resp.status_code == 200


@pytest.mark.nhsd_apim_authorization(access="application", level="level3")
def test_trace(nhsd_apim_proxy_url, nhsd_apim_auth_headers, trace):
    session_name = str(uuid4())
    header_filters = {"trace_id": session_name}
    trace.post_debugsession(session=session_name, header_filters=header_filters)

    resp = requests.get(
        nhsd_apim_proxy_url + "/test-auth/app/level3",
        headers={**header_filters, **nhsd_apim_auth_headers},
    )
    assert resp.status_code == 200

    trace_ids = trace.get_transaction_data(session_name=session_name)

    trace_data = trace.get_transaction_data_by_id(
        session_name=session_name, transaction_id=trace_ids[0]
    )
    status_code_from_trace = trace.get_apigee_variable_from_trace(
        name="message.status.code", data=trace_data
    )

    trace.delete_debugsession_by_name(session_name)

    assert status_code_from_trace == "200"


from typing import Literal, Dict, Optional
import json
import base64
from functools import lru_cache

import pytest

from authlib.jose import jwk
from Crypto.PublicKey import RSA

from .log import log_method
from .token_cache import cache_tokens

from .identity_service import (
    ClientCredentialsConfig,
    ClientCredentialsAuthenticator,
    AuthorizationCodeConfig,
    AuthorizationCodeAuthenticator,
    TokenExchangeConfig,
    TokenExchangeAuthenticator,
    KeycloakUserConfig,
    KeycloakUserAuthenticator,
)


@log_method
@cache_tokens
def get_access_token_via_user_restricted_flow_separate_auth(
    identity_service_base_url,
    keycloak_client_credentials,
    login_form,
    apigee_client_id,
    jwt_private_key,
    jwt_kid,
    auth_scope: Literal["nhs-login", "nhs-cis2"],
    apigee_environment,
):
    if auth_scope == "nhs-cis2":
        # Get token from keycloak
        config = KeycloakUserConfig(
            realm=f"Cis2-mock-{apigee_environment}",
            client_id=keycloak_client_credentials["cis2"]["client_id"],
            client_secret=keycloak_client_credentials["cis2"]["client_secret"],
            login_form=login_form,
        )
        authenticator = KeycloakUserAuthenticator(config=config)
        id_token = authenticator.get_token()["id_token"]

    else:
        # Get token from keycloak
        config = KeycloakUserConfig(
            realm=f"NHS-Login-mock-{apigee_environment}",
            client_id=keycloak_client_credentials["nhs-login"]["client_id"],
            client_secret=keycloak_client_credentials["nhs-login"]["client_secret"],
            login_form=login_form,
        )
        authenticator = KeycloakUserAuthenticator(config=config)
        id_token = authenticator.get_token()["id_token"]

    # Exchange token
    config = TokenExchangeConfig(
        environment=apigee_environment,
        identity_service_base_url=identity_service_base_url,
        client_id=apigee_client_id,
        jwt_private_key=jwt_private_key,
        jwt_kid=jwt_kid,
        id_token=id_token,
    )
    authenticator = TokenExchangeAuthenticator(config=config)
    return authenticator.get_token()


@log_method
@cache_tokens
def get_access_token_via_user_restricted_flow_combined_auth(
    identity_service_base_url: str,
    client_id: str,
    client_secret: str,
    callback_url: str,
    auth_scope: Literal["nhs-login", "nhs-cis2"],
    login_form: Dict[str, str],
    apigee_environment,
):
    config = AuthorizationCodeConfig(
        environment=apigee_environment,
        callback_url=callback_url,
        identity_service_base_url=identity_service_base_url,
        client_id=client_id,
        client_secret=client_secret,
        scope=auth_scope,
        login_form=login_form,
    )
    authenticator = AuthorizationCodeAuthenticator(config=config)
    return authenticator.get_token()


@log_method
@cache_tokens
def get_access_token_via_signed_jwt_flow(
    identity_service_base_url: str,
    client_id: str,
    jwt_private_key: str,
    jwt_kid: str,
    apigee_environment,
):
    config = ClientCredentialsConfig(
        environment=apigee_environment,
        identity_service_base_url=identity_service_base_url,
        client_id=client_id,
        jwt_private_key=jwt_private_key,
        jwt_kid=jwt_kid,
    )
    authenticator = ClientCredentialsAuthenticator(config=config)
    return authenticator.get_token()


@lru_cache(maxsize=None)
@log_method
def create_jwt_key_pair(key_id):
    """
    Pure python instructions to generate a public-key/ private-key
    pair in the correct format.
    """
    key_size = 4096
    key = RSA.generate(key_size)
    private_key_pem = key.exportKey("PEM").decode()
    public_key_pem = key.publickey().exportKey("PEM").decode()

    # This is the JSON formatted public key
    json_web_key = jwk.dumps(
        public_key_pem, kty="RSA", crv_or_size=key_size, alg="RS512"
    )
    json_web_key["kid"] = key_id
    json_web_key["use"] = "sig"

    return {
        "public_key_pem": public_key_pem,
        "private_key_pem": private_key_pem,
        "json_web_key": {"keys": [json_web_key]},
    }


@pytest.fixture(scope="session")
@log_method
def jwt_public_key_id(nhsd_apim_config):
    return nhsd_apim_config["JWT_PUBLIC_KEY_ID"]


@pytest.fixture(scope="session")
@log_method
def _jwt_keys(jwt_public_key_id):
    return create_jwt_key_pair(jwt_public_key_id)


@pytest.fixture(scope="session")
@log_method
def jwt_private_key_pem(_jwt_keys):
    return _jwt_keys["private_key_pem"]


@pytest.fixture(scope="session")
@log_method
def jwt_public_key_pem(_jwt_keys):
    return _jwt_keys["public_key_pem"]


@pytest.fixture(scope="session")
@log_method
def jwt_public_key(_jwt_keys):
    """
    The JWT public key in JSON Web Key format.
    """
    return _jwt_keys["json_web_key"]


@pytest.fixture(scope="session")
@log_method
def jwt_public_key_url(jwt_public_key):
    jwt_public_key_string = json.dumps(jwt_public_key)
    encoded_public_key_bytes = base64.urlsafe_b64encode(jwt_public_key_string.encode())
    return f"https://internal-dev.api.service.nhs.uk/mock-jwks/{encoded_public_key_bytes.decode()}"



"""
Fixtures which interact with Apigee Edge API.

This includes app setup/teardown, getting proxy info (proxy-under-test
+ identity-service of choice), getting products, registering them with
the test app.
"""
import functools
import warnings
from datetime import datetime
from typing import Callable
from uuid import uuid4

import pytest
import requests

from .log import log, log_method
from .apigee_apis import (
    ApigeeNonProdCredentials,
    ApigeeClient,
    AppKeysAPI,
    DebugSessionsAPI,
    AccessTokensAPI,
    ApiProductsAPI,
    DeveloperAppsAPI
)

APIGEE_BASE_URL = "https://api.enterprise.apigee.com/v1/"


@pytest.fixture(scope="session")
@log_method
def _apigee_edge_session(nhsd_apim_config):
    """
    A `requests` session with the correct auth header.
    """
    token = nhsd_apim_config["APIGEE_ACCESS_TOKEN"]
    session = requests.session()
    session.headers = {"Authorization": f"Bearer {token}"}
    return session


@pytest.fixture(scope="session")
@log_method
def _apigee_app_base_url(nhsd_apim_config):
    org = nhsd_apim_config["APIGEE_ORGANIZATION"]
    dev = nhsd_apim_config["APIGEE_DEVELOPER"]
    url = APIGEE_BASE_URL + f"organizations/{org}/developers/{dev}/apps"
    return url

@pytest.fixture(scope="session")
@log_method
def _apigee_app_base_url_no_dev(nhsd_apim_config):
    org = nhsd_apim_config["APIGEE_ORGANIZATION"]
    url = APIGEE_BASE_URL + f"organizations/{org}/apps"
    return url

@functools.lru_cache(maxsize=None)
@log_method
def _get_proxy_json(session, nhsd_apim_proxy_url):
    """
    Query the apigee edge API to get data about the desired proxy, in particular its current deployment.
    """
    deployment_err_msg = (
        "\n\tFailed to retrieve the proxy deployment data. " +
        "Please check the validity of the APIGEE credentials and token as well as any headers."
    )
    deployment_resp = session.get(f"{nhsd_apim_proxy_url}/deployments")
    assert deployment_resp.status_code == 200, deployment_err_msg.format(deployment_resp.content)
    deployment_json = deployment_resp.json()

    # Should be the case
    assert len(deployment_json["environment"]) == 1

    deployed_revision = next(
        filter(
            lambda d: d["state"] == "deployed",
            deployment_json["environment"][0]["revision"],
        )
    )
    revision = deployed_revision["name"]
    proxy_resp = session.get(nhsd_apim_proxy_url + f"/revisions/{revision}")
    assert proxy_resp.status_code == 200
    proxy_json = proxy_resp.json()
    proxy_json["environment"] = deployment_json["environment"][0]["name"]
    return proxy_json


@pytest.fixture()
def _identity_service_proxy(_apigee_edge_session, nhsd_apim_config, _identity_service_proxy_name):
    """
    Get the current revision deployed and pull proxy metadata.
    """
    if not _identity_service_proxy_name:
        return
    org = nhsd_apim_config["APIGEE_ORGANIZATION"]
    url = APIGEE_BASE_URL + f"organizations/{org}/apis/{_identity_service_proxy_name}"
    return _get_proxy_json(_apigee_edge_session, url)


@pytest.fixture()
@log_method
def _apigee_proxy(_apigee_edge_session, nhsd_apim_config, nhsd_apim_proxy_name):
    """
    Get the current revision deployed and pull proxy metadata.
    """
    org = nhsd_apim_config["APIGEE_ORGANIZATION"]
    apigee_edge_api_proxy_url = APIGEE_BASE_URL + f"organizations/{org}/apis/{nhsd_apim_proxy_name}"
    return _get_proxy_json(_apigee_edge_session, apigee_edge_api_proxy_url)


@log_method
def get_all_products(_apigee_edge_session, nhsd_apim_config):
    got_all_products = False
    org = nhsd_apim_config["APIGEE_ORGANIZATION"]
    products_url = APIGEE_BASE_URL + f"organizations/{org}/apiproducts"
    params = {"expand": "true"}
    products = []
    while not got_all_products:
        resp = _apigee_edge_session.get(products_url, params=params)
        new_products = resp.json()["apiProduct"]
        products.extend(new_products)
        if len(new_products) == 1000:
            last = products.pop()
            params.update({"startKey": last["name"]})
        else:
            got_all_products = True
    return products


_APIGEE_PRODUCTS = []


@pytest.fixture()
@log_method
def _proxy_products(_apigee_edge_session, nhsd_apim_proxy_name, nhsd_apim_config):
    """
    Find all products that grant access to your proxy (by name).

    In the case of open-access to an API, ensure this is not called as
    it raises an exception when there are no matches.

    This also allows us to skip checking whether the returned list is
    empty in other fixtures.
    """
    global _APIGEE_PRODUCTS
    proxy_products = [product for product in _APIGEE_PRODUCTS if nhsd_apim_proxy_name in product["proxies"]]

    if len(proxy_products) == 0:
        # Refresh the list and try again...
        _APIGEE_PRODUCTS = get_all_products(_apigee_edge_session, nhsd_apim_config)

    proxy_products = [product for product in _APIGEE_PRODUCTS if nhsd_apim_proxy_name in product["proxies"]]
    if len(proxy_products) == 0:
        raise ValueError(f"No products grant access to proxy {nhsd_apim_proxy_name}")

    return proxy_products


@log_method
def _get_proxy_url(proxy_json):
    """
    Construct the proxy url from the proxy_json
    """
    env = proxy_json["environment"]
    prefix = "https://"
    if env != "prod":
        prefix += f"{env}."
    proxy_url = prefix + "api.service.nhs.uk/" + proxy_json["basepaths"][0]
    return proxy_url


@pytest.fixture()
@log_method
def nhsd_apim_proxy_url(_apigee_proxy):
    """
    Base URL of the proxy under test.
    """
    return _get_proxy_url(_apigee_proxy)


@pytest.fixture()
@log_method
def apigee_environment(_apigee_proxy):
    """
    Apigee environment of the proxy under test.
    """
    return _apigee_proxy["environment"]


@pytest.fixture()
@log_method
def identity_service_base_url(_identity_service_proxy):
    """
    Base URL of the identity-service proxy we will use to authenticate.
    """
    if _identity_service_proxy is None:
        return None
    return _get_proxy_url(_identity_service_proxy)


@pytest.fixture()
@log_method
def _scope(nhsd_apim_authorization):
    """
    Get APIGEE proxy `scope`.
    """
    if not isinstance(nhsd_apim_authorization, type(None)):
        return nhsd_apim_authorization.get("scope")

    return


@pytest.fixture()
@log_method
def _identity_service_proxy_names(_proxy_product_with_scope):
    """
    Get a list of `identity-service` proxies for which we can match a given/required APIGEE `scope`.
    """
    return [proxy for proxy in _proxy_product_with_scope["proxies"] if proxy.startswith("identity-service")]


@pytest.fixture()
@log_method
def _identity_service_proxy_name(_identity_service_proxy_names, nhsd_apim_authorization):
    """
    Make a reasonable choice about which identity-service proxy to use.

    We have a keycloak instance, which is a proper OIDC provider, which is
    pointed to by identity-service proxies with "mock" in the name.

    Return None if there's no identity-service proxies, though this should
    probably never happen.
    """
    if not _identity_service_proxy_names:  # empty list
        return None
    if not nhsd_apim_authorization:
        return None

    keycloak = next(filter(lambda name: "-mock" in name, _identity_service_proxy_names), None)
    if keycloak:
        return keycloak
    warnings.warn(f"Unable to find mock auth generation 2 in {_identity_service_proxy_names}.")
    return _identity_service_proxy_names[0]


@pytest.fixture()
@log_method
def _proxy_product_with_scope(_scope, _proxy_products, nhsd_apim_proxy_name):
    """
    The first product with a scope matching the one specified by the
    pytest.marker.product_scope fixture.

    If the required _scope is None then just returns the first
    product. Otherwise, finds a product that has the required scope.

    Raises an exception if there are no matches.

    This allows us to assume (in other fixtures) that this function
    always returns a product.
    """
    if _scope is None:
        # Any product referencing the proxy under test is fine, so
        # return the first one.
        return _proxy_products[0]
    for product in _proxy_products:
        if _scope in product["scopes"]:
            return product
    error_msg = f"No product granting access to proxy under test has scope `{_scope}`"
    log.error(error_msg)
    raise ValueError(error_msg)


@pytest.fixture(scope="session")
@log_method
def test_app(nhsd_apim_test_app) -> Callable:
    warnings.warn(f"test_app fixture is deprecated. Use nhsd_apim_test_app instead.")
    return nhsd_apim_test_app


_TEST_APP = None


@pytest.fixture(scope="session")
@log_method
def nhsd_apim_test_app(_create_test_app, _apigee_edge_session, _apigee_app_base_url, _apigee_app_base_url_no_dev, _test_app_id) -> Callable:
    """
    A Callable that gets you the current state of the test app.
    """

    # pytest fixtures are wonderful, and do lots of magical things.
    #
    # But...
    #
    # In any well developed pytest extension, one ends up with a
    # complicated dependency graph of fixtures calling other fixtures
    # calling fixtures. In some of these fixtures we want the
    # "current-state" of our app. But some in fixtures we want to
    # update the state of the app. Which fixture gets called first?
    # The answer is... it's really hard to know.
    #
    # As much as I love pytest-provided caching, it's safest to let
    # Apigee be the sole arbiter of the current state of our test app
    # at the cost of an API call.  Therefore, I'm returning a callable
    # rather than JSON from this fixture.
    #
    # If we get the higher level abstractions right in this
    # pytest-extension, a run-of-the-mill user won't need to know much
    # about the app at all, they will just have credentials to call
    # their api.
    def app(force_refresh=False):
        global _TEST_APP
        if _TEST_APP and not force_refresh:
            return _TEST_APP
        if _test_app_id:
            resp = _apigee_edge_session.get(_apigee_app_base_url_no_dev + "/" + _test_app_id)
        else: 
            resp = _apigee_edge_session.get(_apigee_app_base_url + "/" + _create_test_app["name"])
        _TEST_APP = resp.json()
        return _TEST_APP

    return app


@log_method
@pytest.fixture(scope="session")
def nhsd_apim_unsubscribe_test_app_from_all_products(
    nhsd_apim_test_app, _apigee_edge_session, _apigee_app_base_url, _test_app_id
):
    """
    Returns a callable that when run, will unsubscribe the test app
    from all products.

    If you have test code which dynamically creates/destroys products,
    you may need to bring in this callable and execute it.
    """

    def unsubscribe():
        # If app already exists we don't want to modify
        # its product subscriptions
        if _test_app_id:
            return

        app = nhsd_apim_test_app(force_refresh=True)
        app_name = app["name"]
        for cred in app["credentials"]:
            key = cred["consumerKey"]
            resp = _apigee_edge_session.delete(_apigee_app_base_url + f"/{app_name}/keys/{key}")
        app = nhsd_apim_test_app(force_refresh=True)

    return unsubscribe


@log_method
def get_matching_creds(app, product_name):
    """
    Takes some app JSON and gets credentials
    """

    def approved(x):
        return x["status"] == "approved"

    now = int(1000 * datetime.utcnow().timestamp())
    for creds in filter(approved, app["credentials"]):
        if creds["expiresAt"] == -1 or now < creds["expiresAt"]:
            approved_product_names = [p["apiproduct"] for p in filter(approved, creds["apiProducts"])]
            if product_name in approved_product_names:
                return creds


@log_method
def get_app_credentials_for_product(apigee_app_base_url, apigee_edge_session, app, product_name, _test_app_id):
    matching_creds = get_matching_creds(app, product_name)
    if matching_creds is not None:
        return matching_creds

    # If app already exists we do not want to modify its
    # subscriptions.
    if not _test_app_id == "":
        raise ValueError(f"App with id {_test_app_id} does not have expected credentials")

    # Use the apigee edge api to add another set of credentials
    # https://apidocs.apigee.com/docs/developer-apps/1/routes/organizations/%7Borg_name%7D/developers/%7Bdeveloper_email%7D/apps/%7Bapp_name%7D/put
    app["apiProducts"] = [product_name]
    app_url = apigee_app_base_url + "/" + app["name"]
    resp = apigee_edge_session.put(app_url, json=app)
    if resp.status_code != 200:
        raise ValueError(f"Unexpected response from {app_url}: {resp.status_code}, {resp.text}")
    global _TEST_APP
    _TEST_APP = resp.json()

    matching_creds = get_matching_creds(_TEST_APP, product_name)
    return matching_creds


@pytest.fixture()
@log_method
def _test_app_credentials(
    _apigee_app_base_url,
    _apigee_edge_session,
    nhsd_apim_test_app,
    _scope,
    _proxy_product_with_scope,
    _test_app_id,
):
    """
    Get matching credentials for `test_app`, which have access
    to the EXACT set of desired products requested by the user.
    """
    app = nhsd_apim_test_app()
    return get_app_credentials_for_product(
        _apigee_app_base_url,
        _apigee_edge_session,
        app,
        _proxy_product_with_scope["name"],
        _test_app_id,
    )


@pytest.fixture(scope="session")
@log_method
def _apigee_edge_session(nhsd_apim_config):
    """
    Create an APIGEE `requests` session.
    """
    token = nhsd_apim_config["APIGEE_ACCESS_TOKEN"]
    session = requests.session()
    session.headers = {"Authorization": f"Bearer {token}"}
    return session


@pytest.fixture(scope="session")
def nhsd_apim_pre_create_app():
    """
    Hook fixture upon which to hang any methods you wish to call prior
    to creating the test app.
    """
    yield


@pytest.fixture(scope="session")
@log_method
def _create_test_app(
    _apigee_app_base_url,
    _apigee_app_base_url_no_dev,
    _apigee_edge_session,
    jwt_public_key_url,
    nhsd_apim_pre_create_app,
    _test_app_id,
):
    """
    Create an ephemeral app that lasts the duration of the pytest
    session.

    Note that a single app can have many sets of credentials.  Each
    set of credentials can be subscribed to a unique set of products,
    so one app can test your API against multiple product
    configurations should you need to do so.  See `app_credentials`
    for details.
    """

    # Retrieving pre-existing app
    if not _test_app_id == "":
        get_resp = _apigee_edge_session.get(_apigee_app_base_url_no_dev + "/" + _test_app_id)
        err_msg = f"Could not GET TestApp: {_test_app_id}.\tReason: {get_resp.text}"
        assert get_resp.status_code == 200, err_msg
        yield get_resp.json()
    else:
        app = {
            "name": f"apim-auto-{uuid4()}",
            "callbackUrl": "https://example.org/callback",
            "attributes": [{"name": "jwks-resource-url", "value": jwt_public_key_url}],
        }
        create_resp = _apigee_edge_session.post(_apigee_app_base_url, json=app)
        err_msg = f"Could not CREATE TestApp: `{app['name']}`.\tReason: {create_resp.text}"
        assert create_resp.status_code == 201, err_msg

        yield create_resp.json()
        delete_resp = _apigee_edge_session.delete(_apigee_app_base_url + "/" + app["name"])
        err_msg = f"Could not DELETE TestApp: `{app['name']}`.\tReason: {delete_resp.text}"
        assert delete_resp.status_code == 200, err_msg
    global _TEST_APP
    _TEST_APP = None


@pytest.fixture(scope="function")
@log_method
def _create_function_scoped_test_app(
    _apigee_app_base_url,
    _apigee_edge_session,
    jwt_public_key_url,
    nhsd_apim_pre_create_app,
    _test_app_id,
):
    """
    Create an ephemeral app that lasts the duration of the pytest
    test.

    Note that a single app can have many sets of credentials.  Each
    set of credentials can be subscribed to a unique set of products,
    so one app can test your API against multiple product
    configurations should you need to do so.  See `app_credentials`
    for details.
    """

    # Retrieving pre-existing app
    if not _test_app_id == "":
        get_resp = _apigee_edge_session.get(_apigee_app_base_url_no_dev + "/" + _test_app_id)
        err_msg = f"Could not GET TestApp: {_test_app_id}.\tReason: {get_resp.text}"
        assert get_resp.status_code == 200, err_msg
        yield get_resp.json()
    else:
        app = {
            "name": f"apim-auto-{uuid4()}",
            "callbackUrl": "https://example.org/callback",
            "attributes": [{"name": "jwks-resource-url", "value": jwt_public_key_url}],
        }
        create_resp = _apigee_edge_session.post(_apigee_app_base_url, json=app)
        err_msg = f"Could not CREATE TestApp: `{app['name']}`.\tReason: {create_resp.text}"
        assert create_resp.status_code == 201, err_msg

        yield create_resp.json()
        delete_resp = _apigee_edge_session.delete(_apigee_app_base_url + "/" + app["name"])
        err_msg = f"Could not DELETE TestApp: `{app['name']}`.\tReason: {delete_resp.text}"
        assert delete_resp.status_code == 200, err_msg
    global _TEST_APP
    _TEST_APP = None


@pytest.fixture(scope="session")
@log_method
def _test_app_callback_url(_create_test_app):
    return _create_test_app["callbackUrl"]


@pytest.fixture(scope="session")
@log_method
def _test_app_id(nhsd_apim_config):
    return nhsd_apim_config["APIGEE_APP_ID"]

@pytest.fixture()
@log_method
def trace(_apigee_proxy):
    """
    Authenticated wrapper around the DebugSessionsAPI class
    """
    config = ApigeeNonProdCredentials()
    client =  ApigeeClient(config=config)
    debug = DebugSessionsAPI(
        client=client,
        env_name=_apigee_proxy["environment"],
        api_name=_apigee_proxy["name"],
        revision_number=_apigee_proxy["revision"]
    )
    return debug

@pytest.fixture(scope="session")
@log_method
def access_token_api():
    """
    Authenitcated wrapper for Apigee's access token API
    """
    config = ApigeeNonProdCredentials()
    client = ApigeeClient(config=config)
    return AccessTokensAPI(client=client)


@pytest.fixture(scope="session")
@log_method
def products_api():
    """
    Authenitcated wrapper for Apigee's products API
    """
    config = ApigeeNonProdCredentials()
    client = ApigeeClient(config=config)
    return ApiProductsAPI(client=client)

@pytest.fixture(scope="session")
@log_method
def developer_apps_api():
    """
    Authenitcated wrapper for Apigee's developer apps API
    """
    config = ApigeeNonProdCredentials()
    client = ApigeeClient(config=config)
    return DeveloperAppsAPI(client=client)

@pytest.fixture(scope="session")
@log_method
def developer_app_keys_api():
    """
    Authenitcated wrapper for Apigee's developer app keys API
    """
    config = ApigeeNonProdCredentials()
    client = ApigeeClient(config=config)
    return AppKeysAPI(client=client)



"""
This is a self-contained wrapper for a bunch of authentication
methods in APIM. NOT ONLY the identity service is taken into 
account in here, you will also find authenticators for keycloak 
and more... Feel free to keep adding authenticators here and 
maybe move this file to its own library.
"""

import uuid
from abc import ABC, abstractmethod
from time import time
from typing import Literal
from urllib.parse import parse_qs, urlparse

import jwt
import requests
from lxml import html
from pydantic import BaseModel, HttpUrl, validator, AfterValidator
from typing_extensions import Annotated

HttpUrlString = Annotated[HttpUrl, AfterValidator(lambda v: str(v))]

#### Config models
class KeycloakConfig(BaseModel):
    """Basic Keycloak config"""

    realm: Literal[
        "Cis2-mock-internal-dev",
        "Cis2-mock-internal-dev-sandbox",
        "Cis2-mock-internal-qa",
        "Cis2-mock-internal-qa-sandbox",
        "Cis2-mock-ref",
        "Cis2-mock-dev",
        "Cis2-mock-sandbox",
        "Cis2-mock-int",
        "NHS-Login-mock-internal-dev",
        "NHS-Login-mock-internal-dev-sandbox",
        "NHS-Login-mock-internal-qa",
        "NHS-Login-mock-internal-qa-sandbox",
        "NHS-Login-mock-ref",
        "NHS-Login-mock-dev",
        "NHS-Login-mock-sandbox",
        "NHS-Login-mock-int",
        "api-producers",  # just in case u want to get a cheeky token for proxygen ;)
    ]

    @property
    def keycloak_url(self):
        prefix = "https://"
        host = "identity.ptl.api.platform.nhs.uk"
        path = f"/realms/{self.realm}/protocol/openid-connect"
        return f"{prefix}{host}{path}"


class KeycloakUserConfig(KeycloakConfig):
    client_id: str
    client_secret: str
    redirect_uri: HttpUrlString = "https://example.org"
    login_form: dict


class AuthorizationCodeConfig(BaseModel):
    """Config needed to authenticate using authorization_code flow in the identity service"""

    def _identity_service_base_url(env):
        prefix = "https://"
        host = "api.service.nhs.uk"
        path = "/oauth2-mock"
        if env != "prod":
            prefix += f"{env}."
        return f"{prefix}{host}{path}"

    environment: Literal[
        "internal-dev",
        "internal-qa",
        "internal-dev-sandbox",
        "internal-qa-sandbox",
        "ref",
        "sandbox",
        "dev",
        "int",
        "prod",
    ] = "internal-dev"
    org: Literal["nhsd-nonprod", "nhsd-prod"] = "nhsd-nonprod"
    callback_url: HttpUrlString
    identity_service_base_url: HttpUrlString = _identity_service_base_url(environment)
    client_id: str
    client_secret: str
    scope: Literal["nhs-login", "nhs-cis2"]
    login_form: dict

    @validator("environment")
    def validate_environment(cls, environment):
        if environment == "prod":
            raise ValueError(
                f"We dont support testing in the production environment"
            )
        return environment


class ClientCredentialsConfig(BaseModel):
    """Config needed to authenticate using client_credentials flow in the identity service"""

    def _identity_service_base_url(env):
        prefix = "https://"
        host = "api.service.nhs.uk"
        path = "/oauth2-mock"  # lets just support mock auth v2...
        if env != "prod":
            prefix += f"{env}."
        return f"{prefix}{host}{path}"

    environment: Literal[
        "internal-dev",
        "internal-qa",
        "internal-dev-sandbox",
        "internal-qa-sandbox",
        "ref",
        "dev",
        "sandbox",
        "int",
        "prod",
    ] = "internal-dev"
    org: Literal["nhsd-nonprod", "nhsd-prod"] = "nhsd-nonprod"
    client_id: str
    jwt_private_key: str
    jwt_kid: str
    identity_service_base_url: HttpUrlString = _identity_service_base_url(environment)

    def encode_jwt(self):
        url = f"{self.identity_service_base_url}/token"
        claims = {
            "sub": self.client_id,
            "iss": self.client_id,
            "jti": str(uuid.uuid4()),
            "aud": url,
            "exp": int(time()) + 300,  # 5 minutes in the future
        }
        additional_headers = {"kid": self.jwt_kid}
        client_assertion = jwt.encode(
            claims, self.jwt_private_key, algorithm="RS512", headers=additional_headers
        )
        return client_assertion


class TokenExchangeConfig(ClientCredentialsConfig):
    id_token: str


# TODO
class KeycloakSignedJWTConfig:
    pass

# currently only targets AOS environment
class NHSLoginConfig(BaseModel):
    """Config needed to authenticate using NHS Login"""

    def __init__(self, **kwargs):
        openid_config = requests.get("https://auth.aos.signin.nhs.uk/.well-known/openid-configuration").json()
        self.nhs_login_base_url = openid_config["issuer"]

        well_known_jwks: list = requests.get(openid_config["jwks_uri"]).json()
        well_known_key = well_known_jwks['keys'].pop()
        self.jwt_kid = well_known_key["kid"]
        self.alg = well_known_key["alg"]

        super().__init__(**kwargs)

    callback_url: HttpUrlString = "https://nhsd-apim-testing-int-ns.herokuapp.com/nhslogin/callback"
    nhs_login_base_url: HttpUrlString
    client_id: str = "APIM-1"
    jwt_private_key: str
    jwt_kid: str
    alg: str
    scope: str = "openid profile"
    authorize_code: str

    def encode_jwt(self):
        url = f"{self.identity_service_base_url}/token"
        claims = {
            "sub": self.client_id,
            "iss": self.client_id,
            "jti": str(uuid.uuid4()),
            "aud": url,
            "exp": int(time()) + 300,  # 5 minutes in the future
        }
        additional_headers = {"kid": self.jwt_kid}
        client_assertion = jwt.encode(
            claims, self.jwt_private_key, algorithm=self.alg, headers=additional_headers
        )
        return client_assertion
    


class BananaAuthenticatorConfig:  # Placeholder
    pass

### Authenticators
class Authenticator(ABC):
    """Defines the interface"""

    @abstractmethod
    def get_token(self):
        pass


class ClientCredentialsAuthenticator(Authenticator):
    """
    Gets you authenticated using the client_credentials flow in the apim
    identity service, this method of authentication is also known as
    'application restricted' or 'machine user' or 'signed jwt', you can read
    how this pattern works here:
    https://digital.nhs.uk/developer/guides-and-documentation/security-and-authorisation/application-restricted-restful-apis-signed-jwt-authentication
    """

    def __init__(self, config=ClientCredentialsConfig) -> None:
        self.config = config

    def get_token(self):
        login_session = requests.session()
        data = {
            "grant_type": "client_credentials",
            "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            "client_assertion": self.config.encode_jwt(),
        }
        # 1. Do the post call to the identity service
        url = f"{self.config.identity_service_base_url}/token"
        resp = login_session.post(url, data=data)
        # 2. Catch any unexpected error
        if resp.status_code != 200:
            raise RuntimeError(f"{resp.status_code}: {resp.text}")
        # 3. Return your sweet sweet profit
        return resp.json()


class AuthorizationCodeAuthenticator(Authenticator):
    """
    Complete the user-restricted authentication journey for CIS2 or NHS-LOGIN using
    our auth endpoint.

    This webpage does a pretty good job of explaining the journey:
    https://digital.nhs.uk/developer/guides-and-documentation/security-and-authorisation/user-restricted-restful-apis-nhs-cis2-combined-authentication-and-authorisation
    """

    def __init__(self, config=AuthorizationCodeConfig):
        self.config = config

    @staticmethod
    def _get_authorize_endpoint_response(
        session: requests.Session,
        auth_url,
        client_id,
        callback_url,
        auth_scope: Literal["nhs-login", "nhs-cis2"],
    ):
        resp = session.get(
            auth_url,
            params={
                "client_id": client_id,
                "redirect_uri": callback_url,
                "response_type": "code",
                "scope": auth_scope,
                "state": "1234567890",
            },
        )
        if resp.status_code != 200:
            raise RuntimeError(
                f"{auth_url} request returned {resp.status_code}: {resp.text}"
            )
        return resp

    @staticmethod
    def _get_authorize_form_submission_data(authorize_form, login_options):
        inputs = list(authorize_form.inputs)

        form_submission_data = {}
        # This loop picks up the pre-populated defaults, which is
        # sufficient for simulated auth. Defaults can be appended to with
        # the "login_options".
        for _input in inputs:
            input_data = dict(_input.items())
            name = input_data["name"]
            value = input_data["value"]
            form_submission_data[name] = value

        form_submission_data.update(login_options)
        return form_submission_data

    @staticmethod
    def _get_authorization_form(html_str):
        tree = html.fromstring(html_str)
        form = tree.forms[0]
        return form

    @staticmethod
    def _log_in_identity_service_provider(
        session: requests.Session,
        authorize_response,
        authorize_form,
        form_submission_data,
    ):
        form_submit_url = authorize_form.action or authorize_response.request.url
        resp = session.request(
            authorize_form.method, form_submit_url, data=form_submission_data
        )
        # TODO: Investigate why when using the fixtures it returns 404 and when
        # using with external credentials returns 200
        # if resp.status_code != 200:
        #     raise RuntimeError(
        #         f"{form_submit_url} request returned {resp.status_code}: {resp.text}"
        #     )
        return resp

    @staticmethod
    def _get_auth_code_from_mock_auth(response_identity_service_login):
        qs = urlparse(
            response_identity_service_login.history[-1].headers["Location"]
        ).query
        auth_code = parse_qs(qs)["code"]
        if isinstance(auth_code, list):
            # in case there's multiple, this was a bug at one stage
            auth_code = auth_code[0]

        return auth_code

    def get_token(self):
        login_session = requests.session()

        # 1. Hit `authorize` endpoint w/ required query params --> we
        # are redirected to the simulated_auth page. The requests package
        # follows those redirects.
        authorize_response = self._get_authorize_endpoint_response(
            login_session,
            f"{self.config.identity_service_base_url}/authorize",
            self.config.client_id,
            self.config.callback_url,
            self.config.scope,
        )

        authorize_form = self._get_authorization_form(
            authorize_response.content.decode()
        )
        # 2. Parse the login page.  For keycloak this presents an
        # HTML form, which must be filled in with valid data.  The tester
        # can submits their login data with the `login_form` field.

        form_submission_data = self._get_authorize_form_submission_data(
            authorize_form, self.config.login_form
        )

        # form_submission_data["username"] = 656005750104
        #     # And here we inject a valid mock username for keycloak.
        #     # For reference some valid cis2 mock usernames are...
        #     # 656005750104 	surekha.kommareddy@nhs.net
        #     # 656005750105 	darren.mcdrew@nhs.net
        #     # 656005750106 	riley.jones@nhs.net
        #     # 656005750107 	shirley.bryne@nhs.net

        #     # And some valid nhs-login mock usernames are...
        #     # 9912003071      for High - P9
        #     # 9912003072      for Medium - P5
        #     # 9912003073      for Low - P0

        # 3. POST the filled in form. This is equivalent to clicking the
        # "Login" button if we were a human.

        response_identity_service_login = self._log_in_identity_service_provider(
            login_session, authorize_response, authorize_form, form_submission_data
        )
        # 4. The mock auth redirected us back to the
        # identity-service, which redirected us to whatever our app's
        # callback-url was set to.  We don't actually care about the
        # content our callback-url page, we just need the auth_code that
        # was provided in the redirect.
        auth_code = self._get_auth_code_from_mock_auth(response_identity_service_login)

        # 5. Finally, get an access token.
        resp = login_session.post(
            f"{self.config.identity_service_base_url}/token",
            data={
                "grant_type": "authorization_code",
                "code": auth_code,
                "redirect_uri": self.config.callback_url,
                "client_id": self.config.client_id,
                "client_secret": self.config.client_secret,
            },
        )
        resp.raise_for_status()

        # 6. Profit.. sweet sweet profit.
        return resp.json()


class KeycloakUserAuthenticator(Authenticator):
    """
    This is the first step is User Restricted Separate Auth a.k.a Token
    Exchange.  We get a set of tokens from our mock version of CIS2/NHS-LOGIN.
    The reponse includes an ID token, which we will then pass to identity
    service.  Identity service validates the ID *token*, and *exchanges* is it
    for an access token for the NHSD APIM proxies (which probably includes the
    proxy under test).

    It is important to know that for nhs-login we are using client-id and secret
    as supposed to be using signed jwt for authentication as described in the
    documentation. This is so we can use this function to authenticate against
    both providers and at the end of the day the token response from keycloak
    will be the same regardless the authentication method.

    Now if you feel you want to introduce signed-jwt authentication for keycloak
    by all means go for it, hopefully you will find it as easy as implementing a
    new authenticator class...
    """

    def __init__(self, config=KeycloakUserConfig) -> None:
        self.config = config

    def get_token(self):

        login_session = requests.session()
        # 1. Get me that auth page
        resp = login_session.get(
            f"{self.config.keycloak_url}/auth",
            params={
                "response_type": "code",
                "client_id": self.config.client_id,
                "scope": "openid",
                "redirect_uri": self.config.redirect_uri,
            },
        )
        # 2. Parse it!
        tree = html.fromstring(resp.text)
        form = tree.get_element_by_id("kc-form-login")
        # 3. Complete the login form with the credentials in login_form.
        resp2 = login_session.post(form.action, data=self.config.login_form)
        location = urlparse(resp2.history[-1].headers["location"])
        params = parse_qs(location.query)
        # 4. Get me that sweet code from the redirect_uri so I can get my token
        code = params["code"]
        # 5. Get the token
        resp3 = login_session.post(
            f"{self.config.keycloak_url}/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self.config.client_id,
                "client_secret": self.config.client_secret,
                "redirect_uri": self.config.redirect_uri,
            },
        )
        # 6. Return your deserved profit.
        return resp3.json()


class TokenExchangeAuthenticator(Authenticator):
    """
    Get u authenticated using the token_exchange flow in identity service
    """

    def __init__(self, config=TokenExchangeConfig) -> None:
        self.config = config

    # At this point you will be asking yourself 'why we dont dont just implement
    # the hole token exchange flow inside this class?' my answer to that is that
    # I would love every api-producer to understand this flow in detail (you can
    # call me a dreamer) and I believe that forcing them to get a token against
    # Keycloak first and then passing the id_token here makes the hole thing
    # ridiculously explicit. Also how knows... maybe this library will be
    # implemented in some production system one day.
    def get_token(self):
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "subject_token_type": "urn:ietf:params:oauth:token-type:id_token",
            "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            "subject_token": self.config.id_token,
            "client_assertion": self.config.encode_jwt(),
        }
        # 1. Make the post request to the identity service
        url = f"{self.config.identity_service_base_url}/token"
        resp = requests.post(url, data=data)
        # 2. Catch any error
        if resp.status_code != 200:
            raise RuntimeError(f"{resp.status_code}: {resp.text}")
        # 3. Return your profit
        return resp.json()


# TODO:
class KeycloakSignedJWTAuthenticator(Authenticator):
    """Authenticates you against keycloak using signed jwt"""

    def __init__(self, config=KeycloakSignedJWTConfig) -> None:
        self.config = config
        raise NotImplemented(f"TODO")

    def get_token(self):
        raise NotImplemented(f"TODO")


class NHSLoginSandpitAuthenticator(Authenticator):
    """Authenticates you against NHS-Login sandpit environment"""

    def __init__(self, config=NHSLoginConfig) -> None:
        self.config = config
        raise NotImplemented(f"TODO")

    def get_token(self):
        raise NotImplemented(f"TODO")


class NHSLoginAosAuthenticator(Authenticator):
    """Authenticates you against NHS-Login aos environment"""
    # This is only partially implemented. See below for usage:
    # https://nhsd-confluence.digital.nhs.uk/display/APM/KOP-085+Generating+NHS+login+ID+tokens

    def __init__(self, config=NHSLoginConfig):
        super().__init__(config)

    # will need to implement SPA web scraping to do this
    # assuming NHS login don't provide a no-HTML API
    def _get_authorize_code(self):
        raise NotImplemented(f"TODO")

    def get_token(self):
        code = self.config.authorize_code or self._get_authorize_code()
        login_session = requests.session()
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.config.callback_url,
            "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            "client_assertion": self.config.encode_jwt(),
        }
        # 1. Do the post call to the identity service
        url = f"{self.config.nhs_login_base_url}/token"
        resp = login_session.post(url, data=data)
        # 2. Catch any unexpected error
        if resp.status_code != 200:
            raise RuntimeError(f"{resp.status_code}: {resp.text}")
        # 3. Return your sweet sweet profit
        return resp.json()


class NHSLoginProdAuthenticator(Authenticator):
    """Authenticates you against NHS-Login prod environment"""

    def __init__(self, config=NHSLoginConfig) -> None:
        self.config = config
        raise NotImplemented(f"TODO")

    def get_token(self):
        raise NotImplemented(f"TODO")


class BananaAuthenticator(Authenticator):  # Placeholder
    """Authenticates you against a banana"""

    def __init__(self, config=BananaAuthenticatorConfig) -> None:
        self.config = config
        raise NotImplemented(f"TODO")

    def get_token(self):
        raise NotImplemented(f"TODO")



from warnings import warn
from typing import Literal, Union, Dict, Any

import pytest
from typing_extensions import Annotated
from pydantic import BaseModel, Field

from .log import log_method


class BaseAuthorization(BaseModel):
    api_name: str
    force_new_token: bool = False

    def dict(self, **kwargs):
        """
        Construct the product scope when we export the Authorization
        subclasses.
        """
        # Yeah this breaks the inheritance encapsulation
        # slightly... but I'm tired.

        d = super().dict(**kwargs)
        access = d["access"]
        level = d["level"]
        ACCESS_SCOPE_PARTS = {
            "healthcare_worker": "user-nhs-cis2",
            "patient": "user-nhs-login",
            "application": "app",
        }
        scope = ":".join(
            ["urn:nhsd:apim", ACCESS_SCOPE_PARTS[access], level, self.api_name]
        )
        if access == "application" and level == "level0":
            scope = None
        d["scope"] = scope
        return d


class UserRestrictedAuthorization(BaseAuthorization):
    authentication: Literal["combined", "separate"] = "combined"
    login_form: Dict[str, Any] = Field(default_factory=dict)


class HealthcareWorkerAuthorization(UserRestrictedAuthorization):
    """
    Uses CIS2 as the identity provider
    """

    access: Literal["healthcare_worker"]
    level: Literal["aal1", "aal2", "aal3"]



class PatientAuthorization(UserRestrictedAuthorization):
    """
    Uses NHSLogin as the identity provider
    """

    access: Literal["patient"]
    level: Literal["P0", "P5", "P9"]



class ApplicationAuthorization(BaseAuthorization):
    access: Literal["application"]
    level: Literal["level0", "level3"]


class Authorization(BaseModel):
    nhsd_apim_authorization: Annotated[
        Union[
            HealthcareWorkerAuthorization,
            PatientAuthorization,
            ApplicationAuthorization,
        ],
        Field(discriminator="access"),
    ]


@pytest.fixture()
@log_method
def nhsd_apim_authorization(request, nhsd_apim_api_name):
    """
    Mark your test with a `nhsd_apim_authorization marker`.
    Then call the `nhsd_apim_auth_headers` fixture to access your proxy.

    >>> import pytest
    >>> import requests
    >>> @pytest.mark.nhsd_apim_authorization(api_name="mock-jwks", access='healthcare_worker', level="aal3")
    >>> def test_application_restricted_access(nhsd_apim_proxy_url, nhsd_apim_auth_header):
    >>>     resp = requests.get(nhsd_apim_proxy_url + "/a/path/that/is/application/restricted",
    >>>                         headers=nhsd_apim_auth_header)
    >>>     assert resp.status_code == 200
    """
    marker = request.node.get_closest_marker("nhsd_apim_authorization")
    if marker is None:
        warn(
            "Could not find nhsd_apim_authorization marker. This will result in empty authorization headers. Explicitly set an empty @pytest.mark.nhsd_apim_authorization marker() to silence this warning."
        )
        return None

    if not marker.args and not marker.kwargs:
        return None

    if marker.args:
        auth_dict = marker.args[0]
    else:
        auth_dict = dict(**marker.kwargs)

    # Then use the fixture provided value... i.e. the direct marker
    # args override the fixture.
    if "api_name" not in auth_dict:
        auth_dict["api_name"] = nhsd_apim_api_name
    authorization = Authorization(nhsd_apim_authorization=auth_dict)
    return authorization.dict()["nhsd_apim_authorization"]



"""
This module defines the high-level interface to pytest-nhsd-apim.

These are the fixtures that should be used by your typical proxy.
Testers are expected to want a low-friction method of successfully
completing the auth journey so they can check their desired access
patterns.

Test applications register with the actual products they define.

e.g.:

    - /pathA is only allowed with a user-restricted access token.

    - /pathB allows both user-restricted and some flavour of
      application-restricted.
"""
import pytest

# Note: At runtime, pytest does not follow the imports we define in
# our files. Instead, it just looks amongst all the things it found
# when it imported our extension.  This means we have to import *all*
# of our fixtures into this module even if they are only called as
# dependencies of our public fixtures.
from .apigee_edge import (
    _apigee_app_base_url,
    _apigee_app_base_url_no_dev,
    _apigee_edge_session,
    _apigee_proxy,
    _create_function_scoped_test_app,
    _create_test_app,
    _identity_service_proxy,
    _identity_service_proxy_name,
    _identity_service_proxy_names,
    _proxy_product_with_scope,
    _proxy_products,
    _scope,
    _test_app_callback_url,
    _test_app_credentials,
    _test_app_id,
    apigee_environment,
    identity_service_base_url,
    nhsd_apim_pre_create_app,
    nhsd_apim_proxy_url,
    nhsd_apim_test_app,
    nhsd_apim_unsubscribe_test_app_from_all_products,
    test_app,
    trace,
    products_api,
    access_token_api,
    developer_apps_api,
    developer_app_keys_api
)
from .auth_journey import (
    _jwt_keys,
    get_access_token_via_signed_jwt_flow,
    get_access_token_via_user_restricted_flow_combined_auth,
    get_access_token_via_user_restricted_flow_separate_auth,
    jwt_private_key_pem,
    jwt_public_key,
    jwt_public_key_id,
    jwt_public_key_pem,
    jwt_public_key_url,
)

# Import HOOKS so pytest runs them + config fixtures
from .config import nhsd_apim_api_name, nhsd_apim_config, nhsd_apim_proxy_name, pytest_addoption, pytest_configure
from .log import log, log_method
from .nhsd_apim_authorization import nhsd_apim_authorization
from .secrets import (
    _keycloak_client_credentials,
    _mock_jwks_api_key,
    _status_endpoint_api_key,
    status_endpoint_auth_headers,
)


@pytest.fixture()
def _nhsd_apim_auth_token_data(
    nhsd_apim_authorization,
    _test_app_credentials,
    _test_app_callback_url,
    _keycloak_client_credentials,
    identity_service_base_url,
    jwt_private_key_pem,
    jwt_public_key_id,
    apigee_environment,
):
    """
    Main entrypoint to pytest_nhsd_apim.

    This fixture will examine the @pytest.mark.nhsd_apim_authorization
    on your test and do the rest.
    """

    if nhsd_apim_authorization is None:
        return {}

    access = nhsd_apim_authorization["access"]
    level = nhsd_apim_authorization["level"]
    if access == "application":
        if level == "level0":
            return {"apikey": _test_app_credentials["consumerKey"]}
        elif level == "level3":
            token_data = get_access_token_via_signed_jwt_flow(
                identity_service_base_url,
                _test_app_credentials["consumerKey"],
                jwt_private_key_pem,
                jwt_public_key_id,
                apigee_environment,
                force_new_token=nhsd_apim_authorization["force_new_token"],
            )
            return token_data
        else:
            # Should have been pre-validated.
            raise ValueError(f"Invalid level '{level}' for access 'application'.")

    # User restricted auth flows...
    authentication = nhsd_apim_authorization.get("authentication")
    login_form = nhsd_apim_authorization.get("login_form")

    backend_provider_names = {"healthcare_worker": "nhs-cis2", "patient": "nhs-login"}

    if authentication == "combined":
        token_data = get_access_token_via_user_restricted_flow_combined_auth(
            identity_service_base_url,
            _test_app_credentials["consumerKey"],
            _test_app_credentials["consumerSecret"],
            _test_app_callback_url,
            backend_provider_names[access],
            login_form,
            apigee_environment,
            force_new_token=nhsd_apim_authorization["force_new_token"],
        )

    elif authentication == "separate":
        token_data = get_access_token_via_user_restricted_flow_separate_auth(
            identity_service_base_url,
            _keycloak_client_credentials,
            login_form,
            _test_app_credentials["consumerKey"],
            jwt_private_key_pem,
            jwt_public_key_id,
            backend_provider_names[access],
            apigee_environment,
            force_new_token=nhsd_apim_authorization["force_new_token"],
        )
    else:
        raise ValueError(f"Invalid authentication: {authentication}")
    return token_data


@pytest.fixture()
def nhsd_apim_auth_headers(_nhsd_apim_auth_token_data):
    if "access_token" in _nhsd_apim_auth_token_data:
        return {"Authorization": f"Bearer {_nhsd_apim_auth_token_data['access_token']}"}
    elif "apikey" in _nhsd_apim_auth_token_data:
        return _nhsd_apim_auth_token_data
    return {}



"""
pytest-nhsd-apim setup install.
"""

import os

import toml
from setuptools import setup, find_packages


def _read_file(file_name):
    """
    Read a file from disk, given it's `file_name`
    :param file_name: file name, including extension
    :type file_name: str
    :return: file stream
    :rtype: IO stream
    """
    path_to_file = os.path.join(os.path.dirname(__file__), file_name)
    with open(path_to_file) as f_stream:
        return f_stream.read()


def readme():
    """
    Open this project's `README.md`.
    :return: `README.md` file stream
    :rtype: IO stream
    """
    return _read_file("README.md")


def get_pyproject_toml_metadata():
    """
    Read `pyproject.toml`.
    :return: `pyproject.toml` values
    :rtype: dict
    """
    return toml.loads(_read_file("pyproject.toml"))


def get_name_and_email(name_and_email_pair):
    """
    Get name and email from a string of form `FirstName LastName <email_address>`
    :param name_and_email_pair: `FirstName LastName <email_address>`
    :type name_and_email_pair: str
    :return: tuple with name and email address
    :rtype: tuple
    """
    return name_and_email_pair[0].split(' <')[:1][0], name_and_email_pair[0].split(" <")[1:][0][:-1]


def get_package_dependencies(toml_dependencies):
    """
    Read package dependencies from `pyproject.toml`
    :param toml_dependencies: `pyproject.toml` dict containing project dependencies
    :type toml_dependencies: dict
    :return: data needed by setuptools to pass into `install_requires`
    :rtype: list
    """
    install_requires = ["{}=={}".format(pkg_name, pkg_version[1:])
                        for pkg_name, pkg_version in toml_dependencies.items() if pkg_name != 'python']

    return install_requires


PYPROJECT_METADATA = get_pyproject_toml_metadata()
AUTHOR, AUTHOR_EMAIL = get_name_and_email(PYPROJECT_METADATA['tool']['poetry']['authors'])
MAINTAINER, MAINTAINER_EMAIL = get_name_and_email(PYPROJECT_METADATA['tool']['poetry']['maintainers'])

setup(
    name=PYPROJECT_METADATA['tool']['poetry']['name'],
    version=PYPROJECT_METADATA['tool']['poetry']['version'],
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    license=PYPROJECT_METADATA['tool']['poetry']['license'],
    url=PYPROJECT_METADATA['tool']['poetry']['repository'],
    description=PYPROJECT_METADATA['tool']['poetry']['description'],
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"": ["data/*"]},
    python_requires=">=3.8",
    entry_points={"pytest11": ["nhsd_apim = pytest_nhsd_apim.pytest_nhsd_apim"]},
    classifiers=PYPROJECT_METADATA['tool']['poetry']['classifiers'],
    install_requires=get_package_dependencies(PYPROJECT_METADATA['tool']['poetry']['dependencies']),
)



"""
When you're running pytest this module will log into 'pytest' logs.

You can control pytests logging options as described here, so we don't
need to re-invent the wheel.

https://docs.pytest.org/en/7.1.x/how-to/logging.html
"""
import typing
import logging
import inspect
import functools
import uuid
import json
import os
from datetime import datetime

logging.METHOD = 5
logging.addLevelName(logging.METHOD, "METHOD")
logging.Logger.method = lambda inst, msg, *args, **kwargs: inst.log(
    logging.METHOD, msg, *args, **kwargs
)
logging.LoggerAdapter.method = lambda inst, msg, *args, **kwargs: inst.log(
    logging.METHOD, msg, *args, **kwargs
)
logging.method = lambda msg, *args, **kwargs: logging.log(
    logging.METHOD, msg, *args, **kwargs
)

# Use this elswhere
log = logging.getLogger(__name__)


def log_method(f):
    """
    Logs a functions entry and exit.
    Includes args, kwargs, return values, and exceptions.
    Tags each entry/exit pair with a uuid4 for correlation.

    It puts it at a SUPER-LOW logging level (5).
    """

    def pre_log(f, *args, **kwargs):
        log_line = {
            "timestamp": datetime.utcnow().isoformat(),
            "function_name": f.__name__,
            "id": str(uuid.uuid4()),  # use this to match function entry/exit
            "type": "generator" if inspect.isgeneratorfunction(f) else "function",
            "location": "entry",
            "args": list(args),
            "kwargs": dict(**kwargs),
        }
        log.method(log_line)
        return log_line

    def log_and_reraise(log_line, e):
        log_line["exception"] = str(e)
        log.method(log_line)
        raise e

    def post_log(log_line, **extra):
        log_line["location"] = "exit"
        log_line["timestamp"] = datetime.utcnow().isoformat()
        log_line.update(extra)
        log.method(log_line)

    @functools.wraps(f)
    def log_generator(*args, **kwargs):
        log_line = pre_log(f, *args, **kwargs)
        try:
            yield from f(*args, **kwargs)
        except Exception as e:
            log_and_reraise(log_line, e)
        post_log(log_line)

    @functools.wraps(f)
    def log_function(*args, **kwargs):
        log_line = pre_log(f, *args, **kwargs)
        try:
            output = f(*args, **kwargs)
        except Exception as e:
            log_and_reraise(log_line, e)
        post_log(log_line, output=output)
        return output

    return log_generator if inspect.isgeneratorfunction(f) else log_function


def _jsonify(line):
    if isinstance(line, str):
        return line
    try:
        return json.dumps(line, default=str)
    except json.JSONDecodeError:
        return str(line)
